/*******************************************************************************
 * This file is part of ODE Toolkit: a free application for solving systems of
 * ordinary differential equations.
 *  
 * Copyright (C) 2002-2011 Max Comstock, Beky Kotcon, Samantha Mesuro, Daniel Rozenfeld,
 * Anak Yodpinyanee, Andres Perez, Eric Doi, Richard Mehlinger, Steven Ehrlich,
 * Martin Hunt, George Tucker, Peter Scherpelz, Aaron Becker, Eric Harley,
 * Chris Moore
 * 
 * ODE Toolkit is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * ODE Toolkit is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * ODE Toolkit.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/

package data;

import java.awt.geom.Point2D;
import java.util.Collections;
import java.util.Vector;

import solver.SolverParameters;
import util.SolutionReadyListener;

/**
 * Curve is a class for storing all the data generated by one particular set of
 * SolverParameters. A given ODE can have any number of Curves.
 * 
 * @author Clinic 08-09, modified by Clinic 10-11, Max Comstock 2013
 */

public class Curve implements SolutionReadyListener {
	/** Name of the Curve */
	private String name;

	/** 2D Array representing the points on the Curve */
	private double[][] points;

	/** Variables associated to this Curve */
	private ODEVarVector variables = new ODEVarVector();

	/** The stored Bounds for this Curve for avoiding repeated computations */
	private Vector<Bounds> bounds = new Vector<Bounds>();

	/** The number of points presenting in this Curve */
	private int currentNumPoints;
	/** The maximum capacity of points that this Curve can hold */
	private int maxNumPoints;

	/** The state of the solver that generates these points */
	private SolverParameters solverParams;

	/** The visual properties for this Curve */
	private VisualProperties visualProps;

	/**
	 * Constant for double arithmetic error, used in auto-scaling to make sure
	 * floating point error does not result in an incorrect detection of slopes.
	 */
	private static final double ERROR_PADDING = 1.0 + 1E-5;

	/** Minimum fraction of the graph's height to be displayed */
	private static final double MIN_HEIGHT_FRAC = .01;
	/** Minimum fraction of the graph's width to be displayed */
	private static final double MIN_WIDTH_FRAC = .01;

	/**
	 * Constructor that creates a new blank Curve
	 * 
	 * @param params
	 *            SolverParameters used to generate this Curve
	 * @param n
	 *            this Curve's name
	 */
	public Curve(SolverParameters params, String n) {
		name = new String(n);
		solverParams = new SolverParameters(params);
		currentNumPoints = 0;
		maxNumPoints = solverParams.getNumPoints();
		variables = solverParams.getVariables();
		points = null;
		visualProps = new VisualProperties();
	}
	
	/** Constructor that creates a curve with given visual properties
	 * 
	 * @param params
	 *            SolverParameters used to generate this Curve
	 * @param n
	 *            This curve's name
	 * @param vp
	 *            The VisualProperties of this Curve
	 */
	public Curve(SolverParameters params, String n, VisualProperties vp) {
		name = new String(n);
		solverParams = new SolverParameters(params);
		currentNumPoints = 0;
		maxNumPoints = solverParams.getNumPoints();
		variables = solverParams.getVariables();
		points = null;
		visualProps = vp;
	}

	/**
	 * Constructor that creates a Curve with given points and visual properties
	 * 
	 * @param params
	 *            SolverParameters used to generate this Curve
	 * @param n
	 *            this Curve's name
	 * @param p
	 *            the points of this Curve
	 * @param vp
	 *            the VisualProperties of this Curve
	 */
	public Curve(SolverParameters params, String n, double[][] p,
			VisualProperties vp) {
		this(params, n);
		visualProps = vp;
		setPoints(p);
	}

	/**
	 * Constructor that creates a blank Curve with only name field, used only
	 * when loading from file.
	 * 
	 * @param n
	 *            the name of this Curve
	 */
	public Curve(String n) {
		setName(n);
		currentNumPoints = 0;
		maxNumPoints = 0;
		visualProps = new VisualProperties();
	}

	/**
	 * Returns the name of the Curve.
	 * 
	 * @return the name of the Curve
	 */
	public String getName() {
		return name;
	}

	/**
	 * Set the name of the Curve.
	 * 
	 * @param n
	 *            the name for this Curve
	 */
	public void setName(String n) {
		name = new String(n);
	}

	/**
	 * Returns the points of the Curve.
	 * 
	 * @return the points of the Curve
	 */
	public double[][] getPoints() {
		return points;
	}

	/**
	 * Set the points of the Curve.
	 * 
	 * @param p
	 *            the points for this Curve
	 */
	public void setPoints(double[][] p) {
		currentNumPoints = p[0].length;
		maxNumPoints = p[0].length;
		points = p;

		System.out.println("CurveBound Cache Cleared.");

		// clear bounds cache
		bounds.clear();
	}

	/**
	 * Set the points of the Curve, where the given array may contains some
	 * unused space.
	 * 
	 * @param numPoints
	 *            the number of actual points contained in the array
	 * @param p
	 *            the points for this Curve
	 */
	public void setPoints(int numPoints, double[][] p) {
		currentNumPoints = numPoints;
		maxNumPoints = numPoints;
		points = p;

		System.out.println("CurveBound Cache Cleared.");

		// clear bounds cache
		bounds.clear();
	}

	/**
	 * Returns the current number of points the Curve has.
	 * 
	 * @return the current number of points the Curve has
	 */
	public int getCurrentNumPoints() {
		return currentNumPoints;
	}

	/**
	 * Returns the maximum number of points the Curve can hold.
	 * 
	 * @return the maximum number of points the Curve can hold
	 */
	public int getMaxNumPoints() {
		return maxNumPoints;
	}

	/**
	 * Returns the number of dimensions of this Curve.
	 * 
	 * @return the number of dimensions of this Curve
	 */
	public int getDimensions() {
		return variables.size();
	}

	/**
	 * Returns the SolverParameters of the Curve.
	 * 
	 * @return the SolverParameters of the Curve
	 */
	public SolverParameters getSolverParameters() {
		return solverParams;
	}

	/**
	 * Set the SolverParameters of the Curve.
	 * 
	 * @param sp
	 *            the SolverParameters for this Curve
	 */
	public void setSolverParameters(SolverParameters sp) {
		solverParams = sp;
		maxNumPoints = solverParams.getNumPoints();
		variables = solverParams.getVariables();
		currentNumPoints = 0;
	}

	/**
	 * Returns the ODEVarVector containing variables of the Curve.
	 * 
	 * @return the ODEVarVector of the Curve
	 */
	public ODEVarVector getVariables() {
		return variables;
	}

	/**
	 * Determine whether this Curve has a particular variable.
	 * 
	 * @param var
	 *            the ODEVar to check
	 * @return true iff this Curve involves the given variable
	 */
	public boolean hasVar(ODEVar var) {
		return getVariables().contains(var);
	}

	/**
	 * Returns the VisualProperties of this Curve.
	 * 
	 * @return the VisualProperties of this Curve
	 */
	public VisualProperties getVisualProperties() {
		return visualProps;
	}

	/**
	 * Compare whether two Curves are the same.
	 * 
	 * @param otherCurve
	 *            the Curve to be compared with
	 * @return true iff the other Curve is the same as this one
	 */
	public boolean equal(Curve otherCurve) {
		return otherCurve.points == this.points;
	}

	/**
	 * Returns the name of the Curve as printing information.
	 * 
	 * @return the name of the Curve
	 */
	@Override
	public String toString() {
		return getName();
	}

	/**
	 * Implementation of the SolutionReadyListener interface. Sets the points to
	 * the passed point array if this Curve has not already received any
	 * streamed points. If it has, then this function being called means that
	 * the Curve will not receive any more, so it trims the points array to
	 * size.
	 * 
	 * @param sln
	 *            the solution points
	 */
	public void solutionReady(double[][] sln) {
		if (currentNumPoints == 0 && sln != null) {
			if (sln.length == getDimensions()) {
				System.out.println("Points set in solutionReady");
				this.setPoints(sln);
			} else {
				System.out
						.println("ERROR: solution has wrong number of dimensions");
				return;
			}
		}
	}

	/**
	 * Error handling for point and solution ready listeners. Currently, they do
	 * nothing when there is a server error. The ODE that contains this Curve
	 * handles the error situations.
	 * 
	 * @param ex
	 *            the exception to be handled
	 */
	public void errorCondition(Exception ex) {
	}

	/**
	 * Error handling for point and solution ready listeners. Currently, they do
	 * nothing when there is a server error. The ODE that contains this Curve
	 * handles the error situations.
	 * 
	 * @param error
	 *            string describing error condition
	 */
	public void errorCondition(String error) {
	}

	/**
	 * Compute the Bounds for auto-scaling mode
	 * 
	 * @param xVar
	 *            the variable corresponding to x-axis
	 * @param yVar
	 *            the variable corresponding to y-axis
	 * @return the Bounds for auto-scaling mode for this Curve
	 */
	public Bounds computeAutoBounds(ODEVar xVar, ODEVar yVar) {
		if (currentNumPoints == 0) {
			return null;
		}

		int xIndex = variables.indexOf(xVar);
		int yIndex = variables.indexOf(yVar);

		// some variables not present in this Curve
		if (xIndex == -1 || yIndex == -1) {
			Bounds b = new Bounds(xVar, yVar);
			bounds.addElement(b);
			return b;
		}

		for (int i = 0; i < bounds.size(); i++) {

			// found x, y in cache
			if (bounds.elementAt(i).getXVar().equals(xVar)
					&& bounds.elementAt(i).getYVar().equals(yVar)) {
				System.out.println("CurveBound cache hit: " + xVar + ", "
						+ yVar);
				return bounds.elementAt(i);
			}

			// found y, x (switched) in cache, so flip x and y first
			if (bounds.elementAt(i).getXVar().equals(yVar)
					&& bounds.elementAt(i).getYVar().equals(xVar)) {
				System.out.println("CurveBound cache hit: " + xVar + ", "
						+ yVar);

				// switch x with y and y with x
				Bounds b = bounds.elementAt(i);
				Bounds a = b.reverseBound();
				bounds.add(a);

				return a;
			}
		}

		System.out.println("Recalculating CurveBound | cachesize: "
				+ bounds.size() + ", x: " + xVar + ", y:" + yVar);

		// Sort the points in x order.
		// Note: This requires a lot of memory - possible place for improvement
		Vector<Point2D.Double> pts = new Vector<Point2D.Double>();
		Point2D.Double maxY = new Point2D.Double(0, Double.NEGATIVE_INFINITY);
		Point2D.Double minY = new Point2D.Double(0, Double.POSITIVE_INFINITY);
		Point2D.Double maxX = new Point2D.Double(0, Double.NEGATIVE_INFINITY);
		Point2D.Double minX = new Point2D.Double(0, Double.POSITIVE_INFINITY);

		for (int pointIndex = 0; pointIndex < currentNumPoints; pointIndex++) {

			Point2D.Double pp = new Point2D.Double(points[xIndex][pointIndex],
					points[yIndex][pointIndex]);
			pts.add(pp);
			if (pp.y > maxY.y) {
				maxY = pp;
			}
			if (pp.y < minY.y) {
				minY = pp;
			}
			if (pp.x > maxX.x) {
				maxX = pp;
			}
			if (pp.x < minX.x) {
				minX = pp;
			}
		}

		PointComparer pc = new PointComparer();
		Collections.sort(pts, pc);

		int first = 0; // Detect the location of the first element
		// to use for scaling
		boolean beginDecreasing = true; // Detect whether slope is
		// increasing or decreasing.

		// Loop through the points from the beginning and find where
		// an inflection point or a change in direction
		Point2D.Double pt0, pt1, pt2;
		for (int ptIndex = 2; ptIndex < pts.size(); ptIndex++) {
			pt0 = pts.get(ptIndex - 2);
			pt1 = pts.get(ptIndex - 1);
			pt2 = pts.get(ptIndex);

			if (pt1.x != pt0.x) {
				// Detect decreasing slope and same sign (i.e., monotonic)
				if (beginDecreasing) {
					if (Math.abs(slope(pt2, pt1)) <= ERROR_PADDING
							* Math.abs(slope(pt1, pt0))
							&& (pt2.y - pt1.y) * (pt1.y - pt0.y) > 0) {
						first = ptIndex;
					}
					// Swap to detecting for increasing slope, or quit if it
					// has already been decreasing for a period
					else {
						beginDecreasing = false;
						if (ptIndex >= 4)
							break;
					}
				}

				// Detect increasing slope
				else if (Math.abs(slope(pt2, pt1)) >= ERROR_PADDING
						* Math.abs(slope(pt1, pt0))
						&& (pt2.y - pt1.y) * (pt1.y - pt0.y) > 0) {
					first = ptIndex;
				} else
					break;
			}
		}

		int last = pts.size() - 1; // Last element for scaling
		boolean endIncreasing = true; // Detect increasing/dec. slope

		// Loop through the points from the beginning and find where
		// an inflection point or a change in direction
		for (int ptIndex = pts.size() - 3; ptIndex >= 0; ptIndex--) {
			pt0 = pts.get(ptIndex);
			pt1 = pts.get(ptIndex + 1);
			pt2 = pts.get(ptIndex + 2);

			if (pt1.x != pt2.x) {
				if (endIncreasing) {
					// Detect increasing slope and same sign (i.e., monotonic)
					if (Math.abs(slope(pt0, pt1)) <= ERROR_PADDING
							* Math.abs(slope(pt1, pt2))
							&& (pt0.y - pt1.y) * (pt1.y - pt2.y) > 0) {
						last = ptIndex;
					}
					// Swap to detecting for increasing slope, or quit if it
					// has already been decreasing for a period
					else {
						endIncreasing = false;
						if (ptIndex <= pts.size() - 5)
							break;
					}
				}

				// Detect decreasing slope and same sign (i.e., monotonic)
				else if (Math.abs(slope(pt0, pt1)) >= ERROR_PADDING
						* Math.abs(slope(pt1, pt2))
						&& (pt0.y - pt1.y) * (pt1.y - pt2.y) > 0) {
					last = ptIndex;
				} else
					break;
			}
		}

		// If first went past last, we just use the first (initial
		// condition) point plotted.
		double xBottom = points[xIndex][0];
		double xTop = points[xIndex][0];
		double yBottom = points[yIndex][0];
		double yTop = points[yIndex][0];

		if (first <= last) {
			// Otherwise, we look for the bottom and top elements in the
			// entire range
			xBottom = Math.min(pts.elementAt(first).x, pts.elementAt(last).x);
			xTop = Math.max(pts.elementAt(first).x, pts.elementAt(last).x);
			for (int ptIndex = first; ptIndex <= last; ptIndex++) {
				yTop = Math.max(yTop, pts.get(ptIndex).y);
				yBottom = Math.min(yBottom, pts.get(ptIndex).y);
			}
		}

		// but we make sure that we always show at least some of the graph.
		if (yTop - yBottom < MIN_HEIGHT_FRAC * (maxY.y - minY.y)) {
			for (int ptIndex = 0; ptIndex < pts.size(); ptIndex++) {
				if (Math.abs(pts.get(ptIndex).y - points[yIndex][0]) < MIN_HEIGHT_FRAC
						* (maxY.y - minY.y)) {
					xTop = Math.max(xTop, pts.get(ptIndex).x);
					xBottom = Math.min(xBottom, pts.get(ptIndex).x);
					yTop = Math.max(yTop, pts.get(ptIndex).y);
					yBottom = Math.min(yBottom, pts.get(ptIndex).y);
				}
			}
		}
		if (xTop - xBottom < MIN_WIDTH_FRAC * (maxX.x - minX.x)) {
			for (int ptIndex = 0; ptIndex < pts.size(); ptIndex++) {
				if (Math.abs(pts.get(ptIndex).x - points[xIndex][0]) < MIN_WIDTH_FRAC
						* (maxX.x - minX.x)) {
					xTop = Math.max(xTop, pts.get(ptIndex).x);
					xBottom = Math.min(xBottom, pts.get(ptIndex).x);
					yTop = Math.max(yTop, pts.get(ptIndex).y);
					yBottom = Math.min(yBottom, pts.get(ptIndex).y);
				}
			}
		}

		/* Cache the computed Curve bounds */
		Bounds b = new Bounds(xVar, yVar, xBottom, xTop, yBottom, yTop);
		bounds.addElement(b);
		return b;
	}

	/**
	 * Helper function that calculates the slope of a line through pt1 and pt2.
	 * 
	 * @param pt1
	 *            the first point
	 * @param pt2
	 *            the second point
	 * @return slope the slope of line through the given pair of points
	 */
	private static double slope(Point2D.Double pt1, Point2D.Double pt2) {
		return (pt2.y - pt1.y) / (pt2.x - pt1.x);
	}
}
