/*******************************************************************************
 * This file is part of ODE Toolkit: a free application for solving systems of
 * ordinary differential equations.
 *  
 * Copyright (C) 2002-2011 Max Comstock, Beky Kotcon, Samantha Mesuro, Daniel Rozenfeld,
 * Anak Yodpinyanee, Andres Perez, Eric Doi, Richard Mehlinger, Steven Ehrlich,
 * Martin Hunt, George Tucker, Peter Scherpelz, Aaron Becker, Eric Harley,
 * Chris Moore
 * 
 * ODE Toolkit is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * ODE Toolkit is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * ODE Toolkit.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/

package data;

/**
 * Equilibrium is a class for storing all the data generated by one particular
 * initial guess. A given ODE can have any number of Equilibrium.
 * 
 * @author Clinic 10-11
 */

public class Equilibrium {
	/** Name of the Equilibrium */
	private String name;

	/** Array representing the point on the Equilibrium */
	private double[] point;

	/** Variables associated to this Equilibrium */
	private ODEVarVector variables = new ODEVarVector();

	/** The visual properties for this Equilibrium */
	private VisualProperties visualProps;

	/**
	 * Constructor that creates a new Equilibrium
	 * 
	 * @param vars
	 *            the variables associated with this Equilibrium
	 * @param n
	 *            this Equilibrium name
	 * @param p
	 *            the point of this Equilibrium
	 * @param vp
	 *            the VisualProperties of this Equilibrium
	 */
	public Equilibrium(ODEVarVector vars, String n, double[] p,
			VisualProperties vp) {
		name = new String(n);
		variables = vars;
		point = new double[getDimensions()];
		visualProps = vp;
		point = p;
	}

	/**
	 * Constructor that creates a new Equilibrium with default VisualProperties
	 * 
	 * @param vars
	 *            the variables associated with this Equilibrium
	 * @param n
	 *            this Equilibrium name
	 * @param p
	 *            the point of this Equilibrium
	 */
	public Equilibrium(ODEVarVector vars, String n, double[] p) {
		this(vars, n, p, new VisualProperties());
	}

	/**
	 * Returns the name of the Equilibrium.
	 * 
	 * @return the name of the Equilibrium
	 */
	public String getName() {
		return name;
	}

	/**
	 * Set the name of the Equilibrium.
	 * 
	 * @param n
	 *            the name for this Equilibrium
	 */
	public void setName(String n) {
		name = new String(n);
	}

	/**
	 * Returns the point of the Equilibrium.
	 * 
	 * @return the point of the Equilibrium
	 */
	public double[] getPoint() {
		return point;
	}

	/**
	 * Returns the number of dimensions of this Equilibrium.
	 * 
	 * @return the number of dimensions of this Equilibrium
	 */
	public int getDimensions() {
		return variables.size();
	}

	/**
	 * Returns the ODEVarVector containing variables of the Equilibrium.
	 * 
	 * @return the ODEVarVector of the Equilibrium
	 */
	public ODEVarVector getVariables() {
		return variables;
	}

	/**
	 * Determine whether this Equilibrium has a particular variable.
	 * 
	 * @param var
	 *            the ODEVar to check
	 * @return true iff this Equilibrium involves the given variable
	 */
	public boolean hasVar(ODEVar var) {
		return getVariables().contains(var);
	}

	/**
	 * Returns the VisualProperties of this Equilibrium.
	 * 
	 * @return the VisualProperties of this Equilibrium
	 */
	public VisualProperties getVisualProperties() {
		return visualProps;
	}

	/**
	 * Compare whether two Equilibriums are the same.
	 * 
	 * @param otherPoint
	 *            the Equilibrium to be compared with
	 * @return true iff the other Equilibrium is the same as this one
	 */
	public boolean equal(Equilibrium otherPoint) {
		return otherPoint.point == this.point;
	}

	/**
	 * Returns the name of the Equilibrium as printing information.
	 * 
	 * @return the name of the Equilibrium
	 */
	@Override
	public String toString() {
		return getName();
	}

	/**
	 * Compute the Bounds for auto-scaling mode
	 * 
	 * @param xVar
	 *            the variable corresponding to x-axis
	 * @param yVar
	 *            the variable corresponding to y-axis
	 * @return the Bounds for auto-scaling mode for this Equilibrium
	 */
	public Bounds computeImportantPoints(ODEVar xVar, ODEVar yVar) {
		int xIndex = variables.indexOf(xVar);
		int yIndex = variables.indexOf(yVar);

		if (xIndex >= 0 && yIndex >= 0) {
			return new Bounds(xVar, yVar, point[xIndex], point[xIndex],
					point[yIndex], point[yIndex]);
		}
		return new Bounds(xVar, yVar);
	}
}
